module HarmTrace.HarmTrace where

import HarmTrace.Model.Model hiding (PD, PT)
import HarmTrace.Model.Main
import HarmTrace.HAnTree.ToHAnTree
import HarmTrace.HAnTree.HAn
import HarmTrace.HAnTree.Tree
import HarmTrace.HAnTree.Augment
import HarmTrace.Base.MusicRep
import HarmTrace.Tokenizer.Tokens
import HarmTrace.Tokenizer.Tokenizer

-- Parser stuff
import Text.ParserCombinators.UU
import Text.ParserCombinators.UU.BasicInstances as PC

--------------------------------------------------------------------------------
-- Plugging everything together
--------------------------------------------------------------------------------

-- instance Show (Piece key) where show x = showChord x ""

-- parses s with string2PieceC and merges the deleted chords with the tree
string2PieceCPostProc :: String
                      -> ( [ChordLabel], [Piece key], [Tree HAn]
                         , [Error Int], [Error Int])
string2PieceCPostProc s = case string2PieceC s of
  (key, tok, trees, err, err2) -> (tok, trees, output, err, err2) where
    output        = map postProc trees
    postProc tree = treeOperations (mergeDelChords key
                                (groupNeighbours (filterErrorPos err2 tok)) 
                                (gTree $ tree))

{-expandChordDurations :: Tree HAn -> Tree HAn
expandChordDurations (Node h cs a) = (Node (setDur h d) cs' a)  where
     cs' = map expandChordDurations cs
     d   = if null cs' then 0 else maximum $ map (getDur . getLabel) cs'
  -}                              
-- removes some nodes from the tree structure that are not important for
-- similarity estimation
treeOperations :: [Tree HAn] -> Tree HAn
treeOperations = {-expandChordDurations . head . removeInserstions . -}
                 head . removeBy (\l -> l `elem` remv) . head 
  where remv  = [(HAnFunc PD), (HAnFunc PT)]                               
  
{- 
-- do not yet remove this function, allthough it is currently not used (and
-- extremely ugly) we might need it for matching
removeInserstions :: Tree HAn -> [Tree HAn]  
removeInserstions n@(Node l c pn) 
  | isInsertion n  = concatMap removeInserstions c 
  | otherwise = [(Node l (concatMap removeInserstions c) pn)] where 
      isInsertion :: Tree HAn -> Bool  
      isInsertion (Node _ [Node _ [Node _ 
        [Node (HAnChord (ChordToken _ _ _ CT.Inserted _ _)) _ _] _] _] _) = True
      isInsertion (Node _ [Node _ 
        [Node (HAnChord (ChordToken _ _ _ CT.Inserted _ _)) _ _]  _] _) = True
      isInsertion (Node _ 
        [Node (HAnChord (ChordToken _ _ _ CT.Inserted _ _)) _ _ ] _) = True
      isInsertion 
        (Node (HAnChord (ChordToken _ _ _ CT.Inserted _ _)) _ _) = True
      isInsertion  _ = False
-}
      
-- Takes a string with line-separated chords of a song and
-- returns all possible parsed pieces, together with error-correction steps
-- taken (on tokenizing and on musical recognition).
string2PieceC ::String -> (Key,[ChordLabel],[Piece key],[Error Int],[Error Int])
string2PieceC s = let
  (PieceToken key tok,err)=parse ((,) <$> parseSongAbs <*> pEnd) (createStr 0 s)
  (trees,err2) = parse_h ((,) <$> pMajOrMin key <*> pEnd )
                         (createStr 0 (mergeDups key tok))
  in (key, tok, trees, err, err2)
